@@ -4,6 +4,7 @@ from sseclient import SSEClient
 from threading import Thread
 import colorsys
 import os
+from typing import Any, List, Dict, Tuple, Union, Callable, Optional
 
 # Preset colours
 RED = (255, 0, 0)
@@ -25,7 +26,7 @@ class Nanoleaf():
     :ivar print_errors: True for errors to be shown, otherwise False
     """
 
-    def __init__(self, ip, print_errors=False, auth_token=None):
+    def __init__(self, ip : str, auth_token : str =None, print_errors : bool =False):
         """Initalises Nanoleaf class with desired arguments.
 
         :param ip: The IP address of the Nanoleaf device
@@ -51,7 +52,7 @@ class Nanoleaf():
         self.already_registered = False
 
 
-    def __error_check(self, code):
+    def __error_check(self, code : int) -> bool:
         """Checks and displays error messages
 
         Determines the request status code and prints the error, if print_errors
@@ -84,7 +85,7 @@ class Nanoleaf():
                 return False
 
 
-    def create_auth_token(self):
+    def create_auth_token(self) -> Union[str, None]:
         """Creates or retrives the device authentication token
 
         The power button on the device should be held for 5-7 seconds, then
@@ -113,7 +114,7 @@ class Nanoleaf():
         return None
 
 
-    def delete_auth_token(self, auth_token=None):
+    def delete_auth_token(self, auth_token : str =None) -> bool:
         """Deletes an authentication token
 
         Deletes an authentication token and the .nanoleaf_token file if it contains the auth token to delete. 
@@ -133,61 +134,84 @@ class Nanoleaf():
             url = "http://" + self.ip + ":16021/api/v1/" + str(self.auth_token)
         else:
             url = "http://" + self.ip + ":16021/api/v1/" + str(auth_token)
-        r = requests.delete(url)
-        return self.__error_check(r.status_code)
+        response = requests.delete(url)
+        return self.__error_check(response.status_code)
 
-    def check_connection(self):
+    def check_connection(self) -> None:
         """Ensures there is a valid connection"""
         try:
             requests.get(self.url, timeout=5)
         except:
             raise NanoleafConnectionError()
 
-    def get_info(self):
+    def get_info(self) -> Dict[str, Any]:
         """Returns a dictionary of device information"""
-        r = requests.get(self.url)
-        return json.loads(r.text)
+        response = requests.get(self.url)
+        return json.loads(response.text)
 
-    def get_name(self):
+    def get_name(self) -> str:
         """Returns the name of the current device"""
         return self.get_info()['name']
 
-    def get_auth_token(self):
+    def get_auth_token(self) -> str:
         """Returns the current auth token"""
         return self.auth_token
 
+    def get_ids(self) -> List[int]:
+        position_data = []
+        device_ids = []
+        info_data = self.get_info()
+
+        if 'panelLayout' in info_data and 'layout' in info_data['panelLayout'] and 'positionData' in info_data['panelLayout']['layout']:
+            position_data = info_data['panelLayout']['layout']['positionData']
+        
+        # process position data
+        for data in position_data:
+            device_ids.append(data['panelId'])
+
+        return device_ids
+
+    def get_custom_base_effect(self, loop : bool =True) -> Dict[str, Any]:
+        base_effect = {
+            'command': 'display',
+            'animType': 'custom',
+            'loop': loop
+        }
+        return base_effect
+    
+
     #######################################################
     ####                    POWER                      ####
     #######################################################
 
-    def power_off(self):
+    def power_off(self) -> bool:
         """Powers off the lights
 
         :returns: True if successful, otherwise False
         """
         data = {"on" : {"value": False}}
-        r = requests.put(self.url + "/state", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def power_on(self):
+    def power_on(self) -> bool:
         """Powers on the lights
 
         :returns: True if successful, otherwise False
         """
         data = {"on" : {"value": True}}
-        r = requests.put(self.url + "/state", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def get_power(self):
+    def get_power(self) -> bool:
         """Returns the power status of the lights
 
         :returns: True if on, False if off
         """
-        r = requests.get(self.url + "/state/on")
-        ans = json.loads(r.text)
+        response = requests.get(self.url + "/state/on")
+        ans = json.loads(response.text)
         return ans['value']
 
-    def toggle_power(self):
+    def toggle_power(self) -> bool:
         """Toggles the lights on/off"""
         if self.get_power():
             return self.power_off()
@@ -198,7 +222,7 @@ class Nanoleaf():
     ####                   COLOUR                      ####
     #######################################################
 
-    def set_color(self, rgb):
+    def set_color(self, rgb : Tuple[int, int, int]) -> bool:
         """Sets the colour of the lights
 
         :param rgb: Tuple in the format (r, g, b)
@@ -206,24 +230,24 @@ class Nanoleaf():
         :returns: True if successful, otherwise False
         """
         hsv_colour = colorsys.rgb_to_hsv(rgb[0]/255, rgb[1]/255, rgb[2]/255)
-        hsv_colour = list(hsv_colour)
-        hsv_colour[0] *= 360
-        hsv_colour[1] *= 100
-        hsv_colour[2] *= 100
-        final_colour = [ int(x) for x in hsv_colour ]
+        hsv_colour_list = list(hsv_colour)
+        hsv_colour_list[0] *= 360
+        hsv_colour_list[1] *= 100
+        hsv_colour_list[2] *= 100
+        final_colour = [ int(x) for x in hsv_colour_list ]
         data = {
                     "hue" : {"value": final_colour[0]},
                     "sat": {"value": final_colour[1]},
                     "brightness": {"value": final_colour[2], "duration": 0}
                 }
-        r = requests.put(self.url + "/state", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
     #######################################################
     ####               ADJUST BRIGHTNESS               ####
     #######################################################
 
-    def set_brightness(self, brightness, duration=0):
+    def set_brightness(self, brightness : int, duration : int =0) -> bool:
         """Sets the brightness of the lights
 
         :param brightness: The required brightness (between 0 and 100)
@@ -234,10 +258,10 @@ class Nanoleaf():
         if brightness > 100 or brightness < 0:
             raise ValueError('Brightness should be between 0 and 100')
         data = {"brightness" : {"value": brightness, "duration": duration}}
-        r = requests.put(self.url + "/state", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def increment_brightness(self, brightness):
+    def increment_brightness(self, brightness : int) -> bool:
         """Increments the brightness of the lights
 
         :param brightness: How much to increment the brightness, can
@@ -246,32 +270,32 @@ class Nanoleaf():
         :returns: True if successful, otherwise False
         """
         data = {"brightness" : {"increment": brightness}}
-        r = requests.put(self.url + "/state", data = json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data = json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def get_brightness(self):
+    def get_brightness(self) -> int:
         """Returns the current brightness value of the lights"""
-        r = requests.get(self.url + "/state/brightness")
-        ans = json.loads(r.text)
+        response = requests.get(self.url + "/state/brightness")
+        ans = json.loads(response.text)
         return ans['value']
 
     #######################################################
     ####                  IDENTIFY                     ####
     #######################################################
 
-    def identify(self):
+    def identify(self) -> bool:
         """Runs the identify sequence on the lights
 
         :returns: True if successful, otherwise False
         """
-        r = requests.put(self.url + "/identify")
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/identify")
+        return self.__error_check(response.status_code)
 
     #######################################################
     ####                    HUE                        ####
     #######################################################
 
-    def set_hue(self, value):
+    def set_hue(self, value : int) -> bool:
         """Sets the hue of the lights
 
         :param value: The required hue (between 0 and 360)
@@ -281,10 +305,10 @@ class Nanoleaf():
         if value > 360 or value < 0:
             raise ValueError('Hue should be between 0 and 360')
         data = {"hue" : {"value" : value}}
-        r = requests.put(self.url + "/state", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def increment_hue(self, value):
+    def increment_hue(self, value : int) -> bool:
         """Increments the hue of the lights
 
         :param value: How much to increment the hue, can also be negative
@@ -292,20 +316,20 @@ class Nanoleaf():
         :returns: True if successful, otherwise False
         """
         data = {"hue" : {"increment" : value}}
-        r = requests.put(self.url + "/state", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def get_hue(self):
+    def get_hue(self) -> int:
         """Returns the current hue value of the lights"""
-        r = requests.get(self.url + "/state/hue")
-        ans = json.loads(r.text)
+        response = requests.get(self.url + "/state/hue")
+        ans = json.loads(response.text)
         return ans['value']
 
     #######################################################
     ####                 SATURATION                    ####
     #######################################################
 
-    def set_saturation(self, value):
+    def set_saturation(self, value : int) -> bool:
         """Sets the saturation of the lights
 
         :param value: The required saturation (between 0 and 100)
@@ -315,10 +339,10 @@ class Nanoleaf():
         if value > 100 or value < 0:
             raise ValueError('Saturation should be between 0 and 100')
         data = {"sat" : {"value" : value}}
-        r = requests.put(self.url + "/state", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def increment_saturation(self, value):
+    def increment_saturation(self, value : int) -> bool:
         """Increments the saturation of the lights
 
         :param brightness: How much to increment the saturation, can also be
@@ -327,20 +351,20 @@ class Nanoleaf():
         :returns: True if successful, otherwise False
         """
         data = {"sat" : {"increment" : value}}
-        r = requests.put(self.url + "/state", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def get_saturation(self):
+    def get_saturation(self) -> int:
         """Returns the current saturation value of the lights"""
-        r = requests.get(self.url + "/state/sat")
-        ans = json.loads(r.text)
+        response = requests.get(self.url + "/state/sat")
+        ans = json.loads(response.text)
         return ans['value']
 
     #######################################################
     ####              COLOUR TEMPERATURE               ####
     #######################################################
 
-    def set_color_temp(self, value):
+    def set_color_temp(self, value : int) -> bool:
         """Sets the white colour temperature of the lights
 
         :param value: The required colour temperature (between 0 and 100)
@@ -350,10 +374,10 @@ class Nanoleaf():
         if value > 6500 or value < 1200:
             raise ValueError('Colour temp should be between 1200 and 6500')
         data = {"ct" : {"value" : value}}
-        r = requests.put(self.url + "/state", json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def increment_color_temp(self, value):
+    def increment_color_temp(self, value : int) -> bool:
         """Sets the white colour temperature of the lights
 
         :param value: How much to increment the colour temperature by, can also
@@ -362,20 +386,20 @@ class Nanoleaf():
         :returns: True if successful, otherwise False
         """
         data = {"ct" : {"increment" : value}}
-        r = requests.put(self.url + "/state", json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/state", json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def get_color_temp(self):
+    def get_color_temp(self) -> int:
         """Returns the current colour temperature of the lights"""
-        r = requests.get(self.url + "/state/ct")
-        ans = json.loads(r.text)
+        response = requests.get(self.url + "/state/ct")
+        ans = json.loads(response.text)
         return ans['value']
 
     #######################################################
     ####                 COLOUR MODE                   ####
     #######################################################
 
-    def get_color_mode(self):
+    def get_color_mode(self) -> str:
         """Returns the colour mode of the lights"""
         response = requests.get(self.url + "/state/colorMode")
         return json.loads(response.text)
@@ -384,7 +408,7 @@ class Nanoleaf():
     ####                   EFFECTS                     ####
     #######################################################
 
-    def get_current_effect(self):
+    def get_current_effect(self) -> str:
         """Returns the currently selected effect
 
         If the name of the effect isn't available, this will return
@@ -392,10 +416,10 @@ class Nanoleaf():
 
         :returns: Name of the effect or type if unavailable.
         """
-        r = requests.get(self.url + "/effects/select")
-        return json.loads(r.text)
+        response = requests.get(self.url + "/effects/select")
+        return json.loads(response.text)
 
-    def set_effect(self, effect_name):
+    def set_effect(self, effect_name : str) -> bool:
         """Sets the effect of the lights
 
         :param effect_name: The name of the effect
@@ -403,51 +427,137 @@ class Nanoleaf():
         :returns: True if successful, otherwise False
         """
         data = {"select": effect_name}
-        r = requests.put(self.url + "/effects", data=json.dumps(data))
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/effects", data=json.dumps(data))
+        return self.__error_check(response.status_code)
 
-    def list_effects(self):
+    def list_effects(self) -> List[str]:
         """Returns a list of available effects"""
-        r = requests.get(self.url + "/effects/effectsList")
-        return json.loads(r.text)
+        response = requests.get(self.url + "/effects/effectsList")
+        return json.loads(response.text)
 
-    def write_effect(self, effect_dict):
+    def write_effect(self, effect_dict : Dict['str', Any]) -> bool:
         """Writes a user-defined effect to the panels
 
         :param effect_dict: The effect dictionary in the format described here: https://forum.nanoleaf.me/docs/openapi#_u2t4jzmkp8nt
+
+        :raises NanoleafEffectCreationError: When invalid effect dictionary is provided.
+
         :returns: True if successful, otherwise False
         """
-        r = requests.put(self.url + "/effects", data=json.dumps({"write": effect_dict}))
-        if r.status_code == 400:
-            print("Invalid effect dictionary!")
-        return self.__error_check(r.status_code)
+        response = requests.put(self.url + "/effects", data=json.dumps({"write": effect_dict}))
+        if response.status_code == 400:
+            raise NanoleafEffectCreationError("Invalid effect dictionary")
+        return self.__error_check(response.status_code)
 
-    def effect_exists(self, effect_name):
+    def effect_exists(self, effect_name : str) -> bool:
         """Verifies whether an effect exists
 
         :param effect_name: Name of the effect to verify
 
         :returns: True if effect exists, otherwise False
         """
-        r = requests.get(self.url + "/effects/effectsList")
-        if effect_name in json.loads(r.text):
+        response = requests.get(self.url + "/effects/effectsList")
+        if effect_name in json.loads(response.text):
             return True
         return False
 
+    def pulsate(self, rgb : Tuple[int, int, int], speed : int = 1) -> bool:
+        """Displays a pulsating effect on the device with two colours
+        
+        :param rgb: A tuple containing the RGB colour to pulsate in the format (r, g, b).
+        :param speed: The speed of the transition between colours in seconds, with a maximum of 1 decimal place.
+
+        :raises NanoleafEffectCreationError: When an invalid rgb value is provided.
+
+        :returns: True if the effect was created and displayed successfully, otherwise False
+        """
+        if len(rgb) != 3:
+            raise NanoleafEffectCreationError("There must be three values in the RGB tuple! E.g., (255, 0, 0)")
+        for colour in rgb:
+            if not isinstance(colour, int):
+                raise NanoleafEffectCreationError("All values in the tuple must be integers! E.g., (255, 0, 0)")
+            if colour < 0 or colour > 255:
+                raise NanoleafEffectCreationError("All values in the tuple must be integers between 0 and 255! E.g., (255, 0, 0)")
+        base_effect = self.get_custom_base_effect()
+        ids = self.get_ids()
+        anim_data = str(len(ids))
+        frame_string = ""
+        for id in ids:
+            frame_string += " {id} 2".format(id=id)
+            r, g, b = rgb[0], rgb[1], rgb[2]
+            frame_string += " {r} {g} {b} 0 {speed} 0 0 0 0 {speed_2}".format(r=r, g=g, b=b, speed=int(speed*10), speed_2=int(speed*10))
+        base_effect['animData'] = anim_data + frame_string
+        return self.write_effect(base_effect)
+
+    def flow(self, rgb_list : List[Tuple[int, int, int]], speed : int = 1) -> bool:
+        """Displays a sequence of specified colours on the device.
+        
+        :param rgb: A list of tuples containing RGB colours to flow between in the format (r, g, b).
+        :param speed: The speed of the transition between colours in seconds, with a maximum of 1 decimal place.
+
+        :raises NanoleafEffectCreationError: When an invalid rgb_list is provided.
+
+        :returns: True if the effect was created and displayed successfully, otherwise False
+        """
+        if len(rgb_list) <= 1:
+            raise NanoleafEffectCreationError("There has to be more than one tuple in the RGB list for this effect! E.g., [(255, 0, 0), (0, 0, 0)]")
+        for tup in rgb_list:
+            if len(tup) != 3:
+                raise NanoleafEffectCreationError("There must be three values in the RGB tuple! E.g., (255, 0, 0)")
+            for colour in tup:
+                if not isinstance(colour, int):
+                    raise NanoleafEffectCreationError("All values in the tuple must be integers! E.g., (255, 0, 0)")
+                if colour < 0 or colour > 255:
+                    raise NanoleafEffectCreationError("All values in the tuple must be integers between 0 and 255! E.g., (255, 0, 0)")
+        base_effect = self.get_custom_base_effect()
+        ids = self.get_ids()
+        anim_data = str(len(ids))
+        frame_string = ""
+        for id in ids:
+            frame_string += " {id} {numFrames}".format(id=id, numFrames=len(rgb_list))
+            for rgb in rgb_list:
+                r, g, b = rgb[0], rgb[1], rgb[2]
+                frame_string += " {r} {g} {b} 0 {speed}".format(r=r, g=g, b=b, speed=int(speed*10))
+        base_effect['animData'] = anim_data + frame_string
+        return self.write_effect(base_effect)
+
+    def spectrum(self, speed : int = 1) -> bool:
+        """Displays a spectrum cycling effect on the device
+        
+        :param speed: The speed of the transition between colours in seconds, with a maximum of 1 decimal place.
+
+        :returns: True if the effect was created and displayed successfully, otherwise False
+        """
+        base_effect = self.get_custom_base_effect()
+        ids = self.get_ids()
+        spectrum_palette = []
+        for hue in range(0, 360, 10):
+            (r, g, b) = colorsys.hsv_to_rgb(hue/360, 1.0, 1.0)
+            spectrum_palette.append((int(255*r), int(255*g), int(255*b)))
+        anim_data = str(len(ids))
+        frame_string = ""
+        for id in ids:
+            frame_string += " {id} {numFrames}".format(id=id, numFrames=len(spectrum_palette))
+            for rgb in spectrum_palette:
+                r, g, b = rgb[0], rgb[1], rgb[2]
+                frame_string += " {r} {g} {b} 0 {speed}".format(r=r, g=g, b=b, speed=int(speed*10))
+        base_effect['animData'] = anim_data + frame_string
+        return self.write_effect(base_effect)        
+
     #######################################################
     ####                  LAYOUT                       ####
     #######################################################
 
-    def get_layout(self):
+    def get_layout(self) -> Dict[str, Any]:
         """Returns the device layout information"""
-        r = requests.get(self.url + "/panelLayout/layout")
-        return json.loads(r.text)
+        response = requests.get(self.url + "/panelLayout/layout")
+        return json.loads(response.text)
 
     #######################################################
     ####                  EVENTS                       ####
     #######################################################
 
-    def register_event(self, func, event_types):
+    def register_event(self, func : Callable[[Dict[str, Any]], Any], event_types : List[int]) -> None:
         """Starts a thread to register and listen for events
 
         Creates an event listener. This method can only be called once per
@@ -470,27 +580,29 @@ class Nanoleaf():
         if len(event_types) > 4 or len(event_types) < 1:
             raise Exception("The number of events to register for must be" +
                 "between 1-4")
-        for e in event_types:
-            if e < 1 or e > 4:
+        for event in event_types:
+            if event < 1 or event > 4:
                 raise Exception("Valid event types must be between 1-4")
         self.already_registered = True
         t = Thread(target=self.__event_listener, args=(func, set(event_types)))
         t.start()
 
-    def __event_listener(self, func, event_types):
+    def __event_listener(self, func : Callable[[Dict[str, Any]], Any], event_types : List[int]) -> Callable[[], Any]:
         """Listens for events and passes event data to the user-defined
         function."""
-        def inner():
+        def inner() -> Callable[[], Any]:
             url = self.url + "/events?id="
-            for e in event_types:
-                url += str(e) + ","
+            for event in event_types:
+                url += str(event) + ","
             try:
                 messages = SSEClient(url[:-1])
             except Exception as e:
-                print("Events stream failed.")
+                if self.print_errors:
+                    print("Events stream failed.")
                 return inner()
             for msg in messages:
                 func(json.loads(str(msg)))
+            return inner()
         return inner()
 
 
@@ -501,7 +613,7 @@ class Nanoleaf():
 class NanoleafRegistrationError(Exception):
     """Raised when an issue during"""
     
-    def __init__(self):
+    def __init__(self) -> None:
         message = "Authentcation token generation failed. Hold the power button on your Nanoleaf device for 5-7 seconds and try again."
         super(Exception, self).__init__(message)
 
@@ -509,6 +621,13 @@ class NanoleafRegistrationError(Exception):
 class NanoleafConnectionError(Exception):
     """Raised when the connection to the Nanoleaf device fails."""
     
-    def __init__(self):
+    def __init__(self) -> None:
         message = "Connection to Nanoleaf device failed. Is this the correct IP?"
+        super(Exception, self).__init__(message)
+
+
+class NanoleafEffectCreationError(Exception):
+    """Raised when one of the custom effects creation has incorrect arguments."""
+    
+    def __init__(self, message : str) -> None:
         super(Exception, self).__init__(message)
\ No newline at end of file
